from   genolearn.metrics.functions import _metrics
import numpy as np

def _base(Y, Y_hat):
    """ base function to compute statistics required for metric functions in .funcs.py """
    stats  = {}
    unique = set(np.unique(Y)) | set(np.unique(Y_hat))
    for u in sorted(unique):
        TP = (Y_hat[...,Y == u] == u).sum(axis = -1)
        TN = (Y_hat[...,Y != u] != u).sum(axis = -1)
        FP = (Y_hat[...,Y != u] == u).sum(axis = -1)
        FN = (Y_hat[...,Y == u] != u).sum(axis = -1)
        stats[u] = TP, TN, FP, FN
    return stats


def _apply(stats, func):
    """ applies func with stats dictionary from _base """
    with np.errstate(divide = 'ignore', invalid = 'ignore'):
        return {key : func(*value) for key, value in stats.items()}

def _func(name, count):

    if callable(name):
        return name

    elif name is None:
        return name

    def weighted_mean(score):
        return np.einsum('n...,n', score, count) / count.sum()

    funcs = dict(mean = np.mean, weighted_mean = weighted_mean)

    if name in funcs:
        return funcs[name]

    raise Exception()

class Metrics():
    """
    Metrics Class
    
    Computes desired metrics given target labels and predictions.

    Parameters
    ----------
        Y : str
            Array with shape (N,) containing target values.

        Y_hat : str
            Array with shape (...,N) containing predicted values.

        *metrics : str
            String of metrics to use. If None, all metrics are computed.
    """
    def __init__(self, Y, Y_hat, *metrics):
        Y, Y_hat = map(np.array, [Y, Y_hat])
        stats    = _base(Y, Y_hat)
        if len(metrics) == 0:
            self._metric = {metric : _apply(stats, func) for metric, func in _metrics.items()}
        else:
            metrics      = set(metrics) | {'count'}
            self._metric = {metric : _apply(stats, _metrics[metric]) for metric in metrics}
        self.count = self._metric.pop('count')
        
    @property
    def _count(self):
        return np.array(list(self.count.values()))

    def __call__(self, *keys, func = None):
        func = _func(func, self._count)
        ret  = None
        if keys:
            if func:
                ret = {key : func(np.array(list(self._metric[key].values()))) for key in keys}
            else:
                ret = {key : self._metric[key] for key in keys}
        elif func:
            ret = {key : func(np.array(list(value.values()))) for key, value in self._metric.items()}
        else:
            ret = self._metric

        if ret:
            if len(ret) == 1:
                return ret[list(ret)[0]]

        return ret

    def __repr__(self):
        if len(self._metric) == 1:
            Key = list(self._metric)[0]
            sub = '{' + ', '.join([f'{key} : {value:.3f}' for key, value in self._metric[Key].items()]) + '}'
            return f'Metrics({Key} : {sub})'
        keys = list(self._metric)
        return f'Metrics({", ".join(keys[:3])}, ...)' if len(keys) > 3 else f'Metrics({", ".join(keys[:3])})'

    def __getitem__(self, name):
        return self._metric[name]



from   genolearn.logger import msg

from   time             import time
from   datetime         import datetime

import psutil
import click
import json
import os
import re


PARAMS = {}

def get_process_memory():
    """ returns current RAM usage by current Python process """
    return psutil.Process(os.getpid()).memory_info().rss

def monitor_RAM():
    global RAM
    RAM = max(RAM, get_process_memory())
    
def set_params(params):
    """ sets the global PARAMS value """
    global PARAMS
    PARAMS = params.copy()

def get_params():
    """ gets the global PARAMS value """
    global PARAMS
    return PARAMS.copy()
    
def create_log(filename, path = '.'):
    """ creates a text log file with total time elapsed and RAM usage """
    global DATETIME, START, RAM, RAMSTART, PARAMS
    dur  = time() - START
    monitor_RAM()
    ram  = (RAM - RAMSTART) / 1024 ** 3 # B -> GB conversion
    m, s = divmod(dur, 60)
    h, m = divmod(m  , 60)
    contents = f'datetime : {DATETIME}\nduration : {h:02.0f}h {m:02.0f}m {s:02.0f}s\nmemory   : {ram:.3f} GB'

    if PARAMS:
        for key, value in PARAMS.items():
            if isinstance(value, str):
                PARAMS[key] = value.replace(os.path.expanduser('~'), '~')
        contents = f'{contents}\n\n{json.dumps(PARAMS, indent = 4)}'

    contents = f'{contents}\n'
    
    if not filename.endswith('.log'):
        filename = f'{filename}.log'
        
    to_txt(contents, os.path.join(path, filename))

def to_txt(obj, path, mode = 'w'):
    """ writes an object to a text file """
    msg(f'writing "{path}"', inline = True)
    with open(path, mode) as f:
        f.write(obj)
    msg(f'written "{path}"')

def to_json(obj, path):
    """ writes an object to a json file """
    msg(f'writing "{path}"', inline = True)
    with open(path, 'w') as f:
        json.dump(obj, f)
    msg(f'written "{path.split(os.path.sep)[-1] if os.path.sep in path else path}"')

def subdir(path, sub, ext = 0):
    _sub = f'{sub}-{ext}' if ext else sub
    if _sub in os.listdir(path):
        return subdir(path, sub, ext + 1)
    return _sub

def _gen_prompt(msg, type, default ,space, default_option, multiple, show_choices):
    if multiple:
        msg = f'{msg}*'
    msg = f'{msg}' + ' ' * space
    if isinstance(type, click.Choice) and show_choices:
        msg = f'{msg} ' + '{' + ', '.join(map(str, type.choices)) + '}'
    if default_option:
       msg = f'{msg} [{default}]' 
    return msg

def _check(value, type):
    if value == 'None':
        value = None
    try:
        type(value)
        return True
    except:
        return False

def _warn(value, type):
    print(f'bad parameter {value} not {type}')

def _convert_range(string):
    nums   = re.findall('(?<=range\()[0-9, ]+', string)[0].replace(',', ' ').split()
    ls     = list(map(str, list(range(*map(int, nums)))))
    string = ','.join(ls)
    return string

def _prompt(msg, type, default, default_option = True, multiple = False, show_choices = False):

    # user input
    value = input(msg + ': ')

    # check default
    if value == '':
        if default_option:
            value = str(default)
        else:
            print('user input required!')
            return _prompt(msg, type, default, default_option, multiple, show_choices)

    # check range
    if 'range' in value:

        # warn and re-input if type is not int
        if (type != click.INT) and not isinstance(type, click.IntRange):
            _warn(value, type)
            return _prompt(msg, type, default, default_option, multiple, show_choices)

        # convert range string to list of integers
        value = _convert_range(value)

    if value == '*':
        value = ','.join(type.choices)

    # detect if multi-value input
    if ',' in value:
        values = re.split(' ?, ?', value)

        # check each value and return if correct inputs
        for value in values:
            check = _check(value, type)
            if not check:
                _warn(value, type)
                return _prompt(msg, type, default, default_option, multiple, show_choices)
        values = [None if value == None else value for value in map(type, values)]
        return values

    # check (single) value and return if correct input
    check = _check(value, type)
    if check:
        if value == 'None':
            value = None
            return value
        return [type(value)] if multiple else type(value)
    _warn(value, type)
    return _prompt(msg, type, default, default_option, multiple, show_choices)
    
def _default(info):
    if isinstance(info['type'], click.Choice) and len(info['type'].choices) == 1:
        return info['type'].choices[0]
    else:
        return info.get('default', None)

def prompt(params):
    prompts = []
    for key, info in params.items():
        default = _default(info)
        show    = info.get('show_choices', False)
        prompt = _gen_prompt(info.get('prompt', key), info['type'], default, 0, 'default' in info or default, info.get('multiple', False), show)
        prompts.append(prompt)
    
    length = max(map(len, prompts))
    config = {}
    for i, (key, info) in enumerate(params.items()):
        if isinstance(info['type'], click.Choice) and len(info['type'].choices) == 1:
            config[key] = info['type'].choices[0]
            continue
        default = _default(info)
        flag    = 'default' in info or default
        value   = _prompt(f'{prompts[i]:{length}s}', info['type'], default, flag, info.get('multiple', False), info.get('show_choices', False))
        config[key] = value
    return config

START    = time()
RAMSTART = get_process_memory()
RAM      = RAMSTART
DATETIME = datetime.now().strftime('%d-%m-%Y %X')


# from genolearn.core.combine            import core_combine
# from genolearn.core.data               import analyse, core_head, core_tail, train_test_split
# from genolearn.core.evaluate           import core_evaluate
# from genolearn.core.feature_importance import core_feature_importance
# from genolearn.core.feature_selection  import feature_selection as _feature_selection
# from genolearn.core.preprocess         import preprocess, preprocess_meta
# from genolearn.core.train              import core_train



def recall(TP, TN, FP, FN):
    return TP / (TP + FN)

def specificity(TP, TN, FP, FN):
    return TN / (TN + FP)

def precision(TP, TN, FP, FN):
    return TP / (TP + FP)

def negative_predictive_value(TP, TN, FP, FN):
    return TN / (TN + FN)

def false_negative_rate(TP, TN, FP, FN):
    return FN / (FN + TP)

def false_positive_rate(TP, TN, FP, FN):
    return FP / (FP + TN)

def false_discovery_rate(TP, TN, FP, FN):
    return FP / (FP + TP)

def false_omission_rate(TP, TN, FP, FN):
    return FN / (FN + TN)

def positive_likelihood_ratio(TP, TN, FP, FN):
    return recall(TP, TN, FP, FN) / false_positive_rate(TP, TN, FP, FN)

def negative_likelihood_ratio(TP, TN, FP, FN):
    return false_negative_rate(TP, TN, FP, FN) / specificity(TP, TN, FP, FN)

def prevalence_threshold(TP, TN, FP, FN):
    fpr = false_positive_rate(TP, TN, FP, FN)
    tpr = recall(TP, TN, FP, FN)
    return (fpr ** .5) / ((fpr ** .5) + (tpr ** .5))

def threat_score(TP, TN, FP, FN):
    return TP / (TP + FN + FP)

def prevalence(TP, TN, FP, FN):
    return (TP + FN) / (TP + TN + FP + FN)

def accuracy(TP, TN, FP, FN):
    return (TP + TN) / (TP + TN + FP + FN)

def balanced_accuracy(TP, TN, FP, FN):
    return (recall(TP, TN, FP, FN) + specificity(TP, TN, FP, FN)) / 2

def f1_score(TP, TN, FP, FN):
    return 2 * TP / (2 * TP + FP + FN)

def phi_coefficient(TP, TN, FP, FN):
    return (TP * TN + FP * FN) / (((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)) ** .5)

def fowlkes_mallows_index(TP, TN, FP, FN):
    return (false_discovery_rate(TP, TN, FP, FN) * recall(TP, TN, FP, FN)) ** .5

def informedness(TP, TN, FP, FN):
    return recall(TP, TN, FP, FN) + specificity(TP, TN, FP, FN) - 1

def markedness(TP, TN, FP, FN):
    return precision(TP, TN, FP, FN) + negative_predictive_value(TP, TN, FP, FN) - 1

def diagnostics_odds_ratio(TP, TN, FP, FN):
    return positive_likelihood_ratio(TP, TN, FP, FN) /  negative_likelihood_ratio(TP, TN, FP, FN)

def count(TP, TN, FP, FN):
    return TP + FN

_metrics = {metric : func for metric, func in locals().items() if not metric.startswith('_')}



from   genolearn.logger import up, clear, print_dict
from   genolearn.utils  import prompt, _prompt, set_params, get_params
from   genolearn        import __version__, ls, working_directory, get_active, listdir

from   shutil           import rmtree

import resource
import inspect
import pandas as pd
import click
import json
import os


""" general utility variables and functions """

# set preprocessed metadata list
metas  = listdir('meta')

# get active working directory and input metadata
active = get_active()

def user_input(text, n):
    """
    Obtains user input and is bounded between [0, n]
    """
    j = input(text)
    if j.isdigit():
        if 0 <= int(j) < n:
            return int(j), False
    return None, True

def append(command):
    with open(os.path.join(working_directory, '.genolearn')) as file:
        log = json.load(file)
        log['history'].append(command)
    with open(os.path.join(working_directory, '.genolearn'), 'w') as file:
        print(json.dumps(log, indent = 4), file = file, end = '')

def check_working_directory():
    if working_directory:
        if os.path.exists(working_directory):
            if '.genolearn' in listdir():
                return True
    return False

PRE = \
f"""
Genolearn ({__version__}) Command Line Interface

GenoLearn is designed to enable researchers to perform Machine Learning on their genome
sequence data such as fsm-lite or unitig files.

See https://genolearn.readthedocs.io for documentation.
""".strip()

if check_working_directory():
    PRE = f'{PRE}\n\nWorking directory: {working_directory.replace(os.path.expanduser("~"), "~")}'

def enum(options, command = '', pre = '', post = 'user input', k = None, back = None):
    """
    Prints enumerated options to for user input, records the user input, then executes an associated function or returns the option key.

    Parameters
    ----------
        options, (list, dict)
            Available options from the user to choose from. If type dict, the values should be a dictionary with optional entries:
                prompt : prompt to display to the user
                info   : information explaining what the prompt is for
                func   : function to execute should the user choose the option
        
        command, str ['']
            Command text to print prior to pre-text.

        pre, str ['']
            Pre-text to print prior to the enumerated options.

        post, str ['user input']
            Text to print post enumerated options.

        k, int [None]
            Truncate enumerated options.

        back, function [None]
            Function to execute if the user selects to go back (None results in main menu).
    """

    # set back to menu if not provided
    if back is None:
        back = menu

    # set options to only contain back initially (to ensure back is the option 0)
    if back == exit:
        _options = dict(exit = dict(info = 'exits GenoLearn', func = back))
    else:
        _options = dict(back = dict(info = 'goes to the previous command', func = back))

    # add in entries provided
    if isinstance(options, list):
        _options.update({option : {} for i, option in enumerate(options) if k is None or i < k})
    else:
        _options.update({key : value for i, (key, value) in enumerate(options.items()) if k is None or i < k})
        
    options = _options

    # number of options (ignoring back)
    n = len(options)

    # spacing for option number
    m = len(f'{n}')

    # spacing for prompt
    c = max(max(len(option.get('prompt', name)) for name, option in options.items()) + 15, 30)

    # print the pre-text
    if command:
        pre = f'{PRE}\n\nCommand: {command}\n\n{pre}'
    else:
        pre = f'{PRE}\n\nSelect a command'
    print(pre, '\n')

    # print enumerated options
    for i, (name, option) in enumerate(options.items()):
        print(f'{i:{m}d}.  {option.get("prompt", name):{c}s}  {option.get("info", "")}')

        # print extra space to seperate back option from the rest of the option
        if i == 0: print()

    # line space between last option and post-text
    print()

    # obtain user input
    j, flag = user_input(f'{post} : ', n)

    # if incorrect, retry
    while flag:
        print(f'{up}{clear}', end = '')
        j, flag = user_input(f'\r{post} : ', n)

    # clear screen
    pre_n = pre.count('\n') + 1
    print(up * (n + pre_n + 4) + clear, end = '')

    # option key
    key = list(options)[j]

    # if func was provided, execute it
    if 'func' in options[key]:
        func = options[key]['func']
        return func(key) if inspect.signature(func).parameters else func() # if takes in an argument, pass the option key

    # if func not provided, return the option key
    return key

def exit():
    """ Exits the Python shell execution """
    quit()

def _reduce(d, limit):
    """ Truncates dictionary values for printing purposes """
    if isinstance(d, dict):
        if len(d) > limit:
            for i, (k, v) in enumerate(d.copy().items(), 1):
                if i < limit:
                    d[k] = _reduce(v, limit)
                else:
                    del d[k]
                    d['...'] = '...'
        else:
            for k, v in d.items():
                d[k] = _reduce(v, limit)
    elif isinstance(d, list):
        if len(d) > limit:
            d = d[:limit - 1] + ['...']
    return d

def reduce(d, limit):
    """ Truncates dictionary values for printing purposes """
    for k, v in d.items():
        d[k] = _reduce(v, limit)
    return d

def read_log(file):
    """ Reads only the dictionary component of a text file """
    with open(os.path.join(working_directory, file)) as f:
        string = f.read()
        log    = json.loads(string[string.index('{'):])
        return log

def check_history(string):
    """ Checks if any previously executed commands starts with {string}"""
    with open(os.path.join(working_directory, '.genolearn')) as f:
        history = json.load(f)['history']
        for line in history:
            if line.startswith(string):
                return True
        return False

def select_train_dir(command, func, back, train_dir_func = None):
    """ Wrapper for first selecting training directory """
    def _select_train_dir():
        train_dirs = os.listdir(os.path.join(working_directory, 'train'))
        common  = lambda train_dir : {'func' : func, 'info' : train_dir_func(train_dir)} if train_dir_func else {'func' : func}
        options = {train_dir : common(train_dir) for train_dir in train_dirs}
        enum(options, command, 'Select a train subdirectory', back = back)
    return _select_train_dir    
    
def detect_feature_importance(train_dir):
    """ Adds the "info" entry when selecting a train_dir for the feature-importance command """
    if 'importance' in listdir(os.path.join('train', train_dir)):
        return '(already exists)'
    return ''

def detect_evaluate(train_dir):
    """ Adds the "info" entry when selecting a train_dir for the evaluate command """
    path = os.path.join(working_directory, 'train', train_dir, 'evaluate')
    if train_dir in listdir('train') and os.path.exists(path):
        ret = []
        for file in listdir(path):
            ret.append(file.replace('.csv', ''))
        return f'({", ".join(ret)})'
    return ''

""" genolearn-setup """

def setup():
    """ Sets the data directory and proceeds to set the metadata file path """
    def _setup_meta(dir):
        csvs    = [csv for csv in os.listdir(dir) if csv.endswith('.csv') or csv.endswith('.txt')]
        if len(csvs) == 0:
            return print(f'no csv files found in {os.path.abspath(dir)}!')
        options = {csv : {'func' : lambda csv : _setup(dir, csv)} for csv in csvs}
        enum(options, 'setup', 'Select metadata csv for setup', back = setup)
    if working_directory == os.path.abspath('.'):
        return print('current directory is already setup')
    dirs = [dir for dir in os.listdir() if os.path.isdir(dir) and not dir.startswith('_') and not dir.startswith('.')]
    options = {'.' : {'func' : _setup_meta, 'info' : '(current directory)'}}
    for dir in dirs:
        options[dir] = {'func' : _setup_meta}
    try:
        enum(options, 'setup', f'Select data directory for setup within {os.path.abspath(".").replace(os.path.expanduser("~"), "~")}', back = exit)
    except KeyboardInterrupt:
        print()

def _setup(data_dir, meta):
    """ Sets the current working directory and metadata file within it """
    df = pd.read_csv(os.path.join(data_dir, meta))
    if len(df.columns) < 2:
        return print(f'"{meta}" does not contain enough columns')
    config = dict(data_dir = data_dir, meta = meta, history = [f'setup ({__version__})'])
    with open('.genolearn', 'w') as file:
        print(json.dumps(config, indent = 4), file = file, end = '')
    print('setup complete in current directory')


""" genolearn-clean """

def clean():
    """ Deletes all GenoLearn generated files upon user confirmation """
    if working_directory and os.path.exists(working_directory):
        option = dict(confirm = dict(func = _clean, info = 'this cannot be undone'))
        try:
            enum(option, 'clean', 'confirm deletion of all GenoLearn generated files in {working_directory.replace(os.path.expanduser("~"), "~")}?', back = exit)
        except:
            print()
    else:
        print('unknown working directory - either cd into working directory then re-execute genolearn-clean or check if directory already clean')

def _clean():
    """ Deletes all GenoLearn generated files """
    os.chdir(working_directory)
    for dir in ['evaluate', 'feature-selection', 'meta', 'model', 'preprocess', 'train']:
        if dir in ls:
            rmtree(dir)
    hidden = '.genolearn'
    if hidden in ls:
        os.remove(hidden)
    path = os.path.join(os.path.dirname(__file__), 'wd')
    with open(path) as f:
        wd = f.read()
    if wd == working_directory:
        os.remove(path)
    print(f'cleaned {working_directory}')

""" genolearn print """

def __print(name, limit = 5):
    """ Prints various files GenoLearn relies on """

    os.chdir(working_directory)

    if ' ' in name:
        name = name[:name.index(' ')]

    if active is None:
        return print('execute "genolearn config create" first')
    locations = ['meta', 'model']
    if name == 'None':
        for i, location in enumerate(locations):
            loc = location.replace(os.path.expanduser('~'), '~')
            if os.path.exists(location):
                print(f'{loc}\n  - ' + '\n  - '.join(os.listdir(location)))
            else:
                command = 'config model' if i else 'preprocess meta'
                print(f'{loc} (does not exist - execute "genolearn {command}" first)')
            if i < 1:
                print()
    if name == 'config':
        print('.genolearn config\n')
        with open('.genolearn') as f:
            log = json.load(f)
            log.pop('history')
            print(json.dumps(log, indent = 4))
    elif name == 'history':
        print('history\n')
        with open('.genolearn') as f:
            log = json.load(f)
            print('\n'.join(log['history']))
    elif 'meta' in name or 'model' in name:
        directory = os.path.dirname(name)
        dirname   = os.path.basename(directory)
        print(os.path.join(dirname, os.path.basename(name)), '\n')
        with open(name) as f:
            config = reduce(json.load(f), limit)
            print(json.dumps(config, indent = 4).replace('"', '').replace('...: ...', '...'))
    else:
        directory = os.path.dirname(name)
        dirname   = os.path.basename(directory)
        print(os.path.join(dirname, os.path.basename(name)), '\n')
        with open(name) as f:
            print(f.read())

def analyse(meta):
    """ Analyse preprocessed metadata for class label distribution and suggested target subset list """
    info  = dict(min_count  = dict(type = click.IntRange(0), default = 10),
                 proportion = dict(type = click.BOOL, default = False))

    print(f'{PRE}\n\nCommand: print metadata analyse\n\nParameters to analyse "{meta}"\n')
    params = dict(meta = meta)
    params.update(prompt(info))

    params['meta'] = os.path.join('meta', params['meta'])
    
    from   genolearn.core.data import analyse
    
    os.chdir(working_directory)
    if os.path.exists(params['meta']):
        analyse(**params)
    else:
        print('execute genolearn preprocess first')

def head(meta, num = 10):
    """ Prints the first NUM rows of meta data """
    print(os.path.join('meta', meta), '\n')
    meta   = os.path.join(working_directory, 'meta', meta)
    from genolearn.core.data import head
    head(meta, num)


def tail(meta, num = 10):
    """ Prints the last NUM rows of meta data """
    print(os.path.join('meta', meta), '\n')
    meta   = os.path.join(working_directory, 'meta', meta)
    from genolearn.core.data import tail
    tail(meta, num)


def sample(meta, num = 10):
    """ Prints random NUM rows of meta data """
    print(os.path.join('meta', meta), '\n')
    meta   = os.path.join(working_directory, 'meta', meta)
    from genolearn.core.data import sample
    sample(meta, num)

def select_meta(command, pre, func):
    """ Wrapper for first selecting metadata """
    def _select_meta():
        options = {meta : {'func' : func} for meta in metas}
        enum(options, command, pre, back = metadata)
    return _select_meta

def metadata():
    """ metadata subcommand for the print command """
    def args(func):
        name    = func.__name__
        command = f'print metadata {name}'
        pre     = 'Select metadata'
        return command, pre, func

    options = {'analyse' : {'info' : 'analyses the metadata'                , 'func' : select_meta(*args(analyse))},
               'head'    : {'info' : 'prints the head of the metadata'      , 'func' : select_meta(*args(head))},
               'tail'    : {'info' : 'prints the tail of the metadata'      , 'func' : select_meta(*args(tail))},
               'sample'  : {'info' : 'prints random entries of the metadata', 'func' : select_meta(*args(sample))}}
    enum(options, 'print metadata', 'Prints metadata information', back = _print)

def _print():
    """ Prints various files GenoLearn relies on """
    func     = lambda command : __print(command)
    options  = dict(history = dict(info = 'history of significant GenoLearn commands', func = func),
                    config  = dict(info = 'current config', func = func))

    if 'meta' in ls:
        options['meta'] = dict(info = 'metadata information', func = metadata)

    if 'preprocess' in ls:               
        key          = os.path.join(working_directory, 'preprocess', 'preprocess.log')
        options[key] = dict(prompt = 'preprocess.log', info = '(preprocess)', func = func)

        if 'combine.log' in listdir('preprocess'):
            key          = os.path.join(working_directory, 'preprocess', 'combine.log')
            options[key] = dict(prompt = 'combine.log', info = '(preprocess)', func = func)

    for subdir in ['meta', 'feature-selection', 'model']:
        for file in listdir(subdir):
            if subdir != 'feature-selection' or file.endswith('.log'):
                key          = os.path.join(working_directory, subdir, file)
                options[key] = dict(prompt = file, info = f'({subdir})', func = func)

    for subdir in listdir('train'):
        for sub in listdir(os.path.join('train', subdir)):
            if sub.endswith('.log'):
                key          = os.path.join(working_directory, 'train', subdir, sub)
                options[key] = dict(prompt = sub, info = f'({os.path.join("train", subdir)})', func = func)

    enum(options, 'print', 'Select option to print', back = menu)

""" genolearn preprocess """

def preprocess_sequence_data():
    """ Select sequential data to preprocess """
    gzs     = [gz for gz in os.listdir(os.path.join(working_directory, active['data_dir'])) if gz.endswith('.gz')]
    if len(gzs) == 0:
        return print('no sequence data (*.gz) files found!')
    options = {}
    for gz in gzs:
        info = ''
        if 'preprocess' in listdir():
            log = read_log(os.path.join('preprocess', 'preprocess.log'))
            if os.path.basename(log['data']) == gz:
                info = '(already preprocessed)'
            if 'combine.log' in listdir('preprocess'):
                log = read_log(os.path.join('preprocess', 'combine.log'))
                if isinstance(log['data'], str):
                    data = [os.path.basename(log['data'])]
                else:
                    data = [os.path.basename(data) for data in log['data']]
                if gz in data:
                    info = '(already combined)'
        options[gz] = {'func' : preprocess_sequence, 'info' : info}
    enum(options, 'preprocess sequence', 'Select option to preprocess', back = preprocess)

def preprocess_sequence(data):
    """ Preprocesses sequential data """
    print(f'{PRE}\n\nparameters for "{data}" to preprocess\n')
    info   = dict(batch_size = dict(type = click.INT, default = None),
                  n_processes = dict(type = click.INT, default = None),
                  sparse = dict(type = click.BOOL, default = True),
                  dense = dict(type = click.BOOL, default = True),
                  verbose = dict(type = click.IntRange(1), default = 250000),
                  max_features = dict(type = click.IntRange(-1), default = None))

    params = dict(data = data)
    params.update(prompt(info))

    assert params['dense'] or params['sparse'], 'set either / both dense and sparse to True'

    params['data']         = os.path.join(working_directory, active['data_dir'], data).replace(os.path.expanduser('~'), '~')

    if params['max_features'] != None:
        params['verbose'] = params['max_features'] // 10

    from multiprocessing import cpu_count

    if params['batch_size'] == None:
        params['batch_size'] = min(resource.getrlimit(resource.RLIMIT_NOFILE)[1], 2 ** 14) # safety
    if params['n_processes'] == None:
        params['n_processes'] = cpu_count()

    print_dict('executing "preprocess sequence" with parameters:', params)

    from   genolearn.core.preprocess import preprocess as core_preprocess

    os.chdir(working_directory)
    core_preprocess('preprocess', **params)
    append(f'preprocess sequence ({data})')

def preprocess_combine_data():
    """ Select sequential data to preprocess and combine with already preprocessed data """
    with open(os.path.join(working_directory, '.genolearn')) as f:
        log = json.load(f)
        preprocessed = []
        for line in log['history']:
            if line.startswith('preprocess sequence'):
                preprocessed.append(line[21:-1])
    if 'combine.log' in listdir('preprocess'):
        log  = read_log(os.path.join('preprocess', 'combine.log'))
        if isinstance(log['data'], str):
            preprocessed.append(os.path.basename(log['data']))
        else:
            preprocessed += [os.path.basename(file) for file in log['data']]
    options = {}
    for data in listdir(active['data_dir']):
        if data.endswith('.gz') and data not in preprocessed:
            options[data] = {'func' : preprocess_combine}
    if len(options) == 0:
        return print('no un-preprocessed sequence data (.gz) files found!')
    enum(options, 'preprocess combine' ,'Select option to preprocess and combine', back = preprocess)

def preprocess_combine(data):
    """ Preprocess sequential data and combine with already preprocessed data """
    print(f'preprocess {data}')
    info = dict(batch_size   = dict(type = click.INT, default = None),
                n_processes  = dict(type = click.INT, default = None),
                verbose      = dict(type = click.INT, default = 250000))

    params = dict(data = data)
    params.update(prompt(info))

    meta   = read_log(os.path.join('preprocess', 'preprocess.log'))
    params['max_features'] = meta['max_features']

    params['data'] = os.path.join(working_directory, active['data_dir'], data).replace(os.path.expanduser('~'), '~')

    if params['max_features'] != None:
        params['verbose'] = params['max_features'] // 10
        
    from multiprocessing import cpu_count

    if params['batch_size'] == None:
        params['batch_size'] = min(resource.getrlimit(resource.RLIMIT_NOFILE)[1], 2 ** 14) # safety
    if params['n_processes'] == None:
        params['n_processes'] = cpu_count()

    print_dict('executing "preprocess combine" with parameters:', params)

    from   genolearn.core.preprocess import combine

    if 'combine.log' in listdir('preprocess'):
        log    = read_log(os.path.join('preprocess', 'combine.log'))
        PARAMS = get_params().copy()
        if isinstance(log['data'], str):
            PARAMS['data']  = [log['data'], params['data'].replace(os.path.expanduser('~'), '~')]
        else:
            PARAMS['data'] += [log['data']]
        set_params(PARAMS)
    os.chdir(working_directory)
    combine('preprocess', **params)
        
    append(f'preprocess combine ({data})')

def preprocess_meta():
    """ Preprocesses metadata """
    meta_path      = os.path.join(working_directory, active['data_dir'], active['meta'])
    print(f'{PRE}\n\nCommand: preprocess meta\n\nParameters for preprocessing "{os.path.basename(active["meta"])}"\n')

    meta_df        = pd.read_csv(meta_path).applymap(str)
    valid_columns  = set(meta_df.columns)
    
    output         = click.prompt('output       ', type = click.STRING, default = 'default')
    identifier     = click.prompt('identifier             ', type = click.Choice(valid_columns), show_choices = False)

    valid_columns -= set([identifier])

    target         = click.prompt('target                 ', type = click.Choice(valid_columns), show_choices = False)

    valid_columns -= set([target])
    valid_columns |= set(['None'])

    group          = click.prompt('group           ', type = click.Choice(valid_columns), show_choices = False, default = 'None')

    if group != 'None':
        groups       = set(sorted(set(meta_df[group])))
        train_values = _prompt('train group values*    ', type = click.Choice(groups), default = None, default_option = False, multiple = True)
        groups      -= set(train_values)
        test_values  = _prompt('test  group values*    ', type = click.Choice(groups), default = None, default_option = False, multiple = True)
        ptrain       = None
    else:
        train_values = ['Train']
        test_values  = ['Test']
        ptrain       = click.prompt('proportion train', type = click.FloatRange(0., 1.), default = 0.75)

    from genolearn.core.preprocess import preprocess_meta

    os.chdir(working_directory)
    preprocess_meta(output, meta_path, identifier, target, group, train_values, test_values, ptrain)
    append(f'preprocess meta ({output})')

def preprocess():
    """ Preprocess command """
    options = {'sequence' : {'info' : 'preprocesses sequence data',
                             'func' : preprocess_sequence_data},
               'combine'  : {'info' : 'preprocesses sequence data and combines to previous preprocessing',
                             'func' : preprocess_combine_data},
               'meta'     : {'info' : 'preprocesses meta data',
                             'func' : preprocess_meta}}
    enum(options, 'preprocess', 'Select a preprocess subcommand', k = None if check_history('preprocess sequence') else 1)

""" genolearn feature-selection """

def _feature_selection():
    """ Wrapper for first selecting metadata and then feature-selection method """

    path = 'feature-selection'

    def detect(meta, method = None):
        ret = []
        if os.path.exists(os.path.join(working_directory, path)):
            for file in listdir(path):
                if file.endswith('.log'):
                    log = read_log(os.path.join(path, file))
                    if method:
                        if (log['meta'], log['method']) == (meta, method):
                            return '(already exists)'
                    elif log['meta'] == meta:
                        ret.append(log['method'])
        return f'({", ".join(ret)})' if ret else ''
        
    def _select_feature_selection(meta):
        fisher  = os.path.join(os.path.dirname(__file__), 'core', 'feature_selection', 'fisher.py')
        func    = lambda method : feature_selection(meta, method)
        exists  = detect(meta, 'fisher')
        options = {fisher : {'prompt' : 'fisher', 'info' : exists if exists else 'Fisher Score for Feature Selection', 'func' : func}}
        for dir in set([working_directory, os.path.abspath('.')]):
            for file in os.listdir(dir):
                if file.endswith('.py'):
                    py  = file.replace('.py', '')
                    key = os.path.join(dir, file)
                    options[key] = {'prompt' : py, 'func' : func, 'info' : detect(meta, py)}        
        enum(options, 'feature-selection', f'Select a feature selection method to use for "{meta}" metadata', back = _feature_selection)

    options = {meta : {'func' : _select_feature_selection, 'info' : detect(meta)} for meta in metas}
    enum(options, 'feature-selection', 'Select a metadata file')
            
def feature_selection(meta, module):
    """ Computes Feature Selection (Fisher by default) """
    print(f'{PRE}\n\nCommand: feature-selection\n\nParameters for feature selection using "{meta}" meta with "{os.path.basename(module)[:-3]}" method\n')
    py     = os.path.basename(module).replace('.py', '')
    params = dict(meta = meta, method = py, module = module.replace(os.path.expanduser('~'), '~'))
    info   = dict(name = dict(default = f'{params["meta"]}-{py}', type = click.STRING))
    
    params.update(prompt(info))

    print_dict('executing "feature-selection" with parameters:', params)

    params['log'] = f"{params['name']}.log"

    from   genolearn.core.feature_selection import feature_selection
    os.chdir(working_directory)
    feature_selection(**params)
    append(f'feature-selection ({params["name"]})')

""" genolearn model-config """

classifiers = dict(
                  logistic_regression = \
                  dict(model = 'LogisticRegression',
                          config_name = dict(type = click.STRING, default = 'logistic-regression'),
                          penalty = dict(type = click.Choice(['l1', 'l2', 'elasticnet', 'none']), default = 'l2', show_choices = True),
                          dual = dict(type = click.BOOL, default = False),
                          tol = dict(type = click.FloatRange(1e-8), default = 1e-4),
                          C = dict(type = click.FloatRange(1e-8), default = 1.),
                          fit_intercept = dict(type = click.BOOL, default = True),
                          class_weight = dict(type = click.Choice(['balanced', 'None']), default = 'None', show_choices = True),
                          random_state = dict(type = click.INT, default = None),
                          solver = dict(type = click.Choice(['newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga']), default = 'lbfgs', show_choices = True),
                          max_iter = dict(type = click.IntRange(1), default = 100),
                          multi_class = dict(type = click.Choice(['auto', 'ovr', 'multinomial']), default = 'auto', show_choices = True),
                          n_jobs = dict(type = click.IntRange(-1), default = -1),
                          l1_ratio = dict(type = click.FloatRange(0), default = 1.)),
                  random_forest = \
                  dict(model = 'RandomForestClassifier',
                          config_name = dict(type = click.STRING, default = 'random-forest'),
                          n_estimators = dict(type = click.IntRange(1), default = 100),
                          criterion = dict(type = click.Choice(['gini', 'entropy', 'log_loss']), default = 'gini', show_choices = True),
                          max_depth = dict(type = click.IntRange(1), default = None),
                          min_samples_split = dict(type = click.IntRange(1), default = 2),
                          min_samples_leaf = dict(type = click.IntRange(1), default = 1),
                          min_weight_fraction_leaf = dict(type = click.FloatRange(0., 0.5), default = 0.),
                          max_features = dict(type = click.Choice(['sqrt', 'log2', 'None']), default = 'sqrt', show_choices = True),
                          max_leaf_nodes = dict(type = click.IntRange(1), default = None),
                          min_impurity_decrease = dict(type = click.FloatRange(0), default = 0.),
                          bootstrap = dict(type = click.BOOL, default = True),
                          oob_score = dict(type = click.BOOL, default = False),
                          n_jobs = dict(type = click.IntRange(-1), default = -1),
                          random_state = dict(type = click.INT, default = None),
                          class_weight = dict(type = click.Choice(['balanced', 'balanced_subsample', 'None']), default = None, show_choices = True))
                  )

def _model(name):
    """ Given a model name, prompts user for hyperparameter settings """
    classifier  = classifiers[name]
    params      = {'model' : classifier.pop('model')}
    print(f'{PRE}\n\nCommand: model-config\n\nParameters for {params["model"]}\n')
    params.update(prompt(classifier))
    config_name = params.pop('config_name')
    path        = os.path.join(working_directory, 'model')
    os.makedirs(path, exist_ok = True)
    with open(os.path.join(path, config_name), 'w') as file:
        print(json.dumps(params, indent = 4), file = file)
    print(f'generated "{config_name}" in {path}')
    append(f'model ({config_name})')

def model_config():
    """ Prompts user for model to use before prompting for hyperparameter settings """
    models  = ['logistic_regression', 'random_forest']
    options = {model : {'func' : _model} for model in models}
    enum(options, 'model-config', 'Select a model to configure')

""" genolearn train """

def detect_train(meta, feature_selection = None, model_config = None):
    """ Adds the "info" entry when selecting a metadata file for the train command """
    ret    = []
    num    = 1 + bool(feature_selection) + bool(model_config)
    target = (meta, feature_selection, model_config)[:num]
    if 'train' in ls:
        for train_dir in listdir('train'):
            log = read_log(os.path.join('train', train_dir, 'train.log'))
            if (log['meta'], log['feature_selection'], log['model_config'])[:num] == target:
                ret.append(train_dir)
    return f'({", ".join(ret)})' if ret else ''

def _train():
    def _select_model_config(meta_feature_selection):
        meta, feature_selection = meta_feature_selection
        options = {}
        for model_config in listdir('model'):
            func = lambda model_config : train(meta, feature_selection, model_config)
            info = detect_train(meta, feature_selection, model_config)
            options[model_config] = {'func' : func, 'info' : info}
        pre_text = f'Select a model config to use with "{meta}" metadata and "{feature_selection}" feature selection to train'
        enum(options, 'train', pre_text, back = _train)
    options = {}
    for selection in listdir('feature-selection'):
        if selection.endswith('.log'):
            log  = read_log(os.path.join('feature-selection', selection))
            meta = log['meta']
            selection = selection.replace('.log', '')
            options[(meta, selection)] = {'prompt' : selection, 'info' : f'("{meta}" metadata)', 'func' : _select_model_config}
    enum(options, 'train', 'Select a feature selection file to use to train')

def train(meta, feature_selection, model_config):
    """ Given a preprocessed metadata file, trains model(s) and save outputs to the train subdirectory within the working directory """
    params = locals().copy()
    print(f'{PRE}\n\nCommand: train\n\nTrain parameters for metadata "{meta}" with feature-selection "{feature_selection}" and model config "{model_config}"\n')

    default = f'{feature_selection}-{model_config}'
    group   = ['Train', 'Test']
    with open(os.path.join(working_directory, 'meta', meta)) as f:
        meta_ = json.load(f)
        if set(meta_['group']) != {'Train', 'Test'}:
            group = list(meta_['group']) + group
        
    from genolearn.metrics import _metrics

    choice  = click.Choice(sorted(set(_metrics) - {'count'}))
    info    = dict(output_dir = dict(type = click.Path(), default = default),
                   num_features = dict(default = 1000, type = click.IntRange(1), multiple = True),
                   min_count = dict(default = 0, type = click.IntRange(0)),
                   target_subset = dict(default = 'None', type = click.Choice(group)),
                   metric = dict(default = 'f1_score', type = choice, show_choices = False),
                   aggregate_func = dict(default = 'weighted_mean', type = click.Choice(['mean', 'weighted_mean'])))

    params.update(prompt(info))

    os.chdir(working_directory)
    os.makedirs('train', exist_ok = True)
    
    params['output_dir'] = os.path.join('train', params['output_dir'])
    
    if os.path.exists(params['output_dir']):
        rmtree(params['output_dir'])

    if isinstance(params['num_features'], int):
        params['num_features'] = [params['num_features']]

    print_dict('executing "genolearn train" with parameters:', params)

    from genolearn.core.train import train
    train(**params)
    append(f'train ({os.path.basename(params["output_dir"])})')

""" genolearn feature-importance """

def feature_importance(train_dir):
    """ Given a training directory, computes the Feature Importance and outputs an Importance subdirectory """
    params = dict(train_dir = train_dir)

    print_dict('executing "genolearn feature-importance" with parameters:', params)
    
    params['train_dir'] = os.path.join('train', params['train_dir'])
    params['model']     = os.path.join(params['train_dir'], 'model.pickle')
    params['output']    = os.path.join(params['train_dir'], 'importance')
    
    os.chdir(working_directory)

    log = read_log(os.path.join(params.pop('train_dir'), 'train.log'))
    params['feature_selection'] = log['feature_selection']
    params['meta'] = log['meta']

    from   genolearn.core.feature_importance import feature_importance

    feature_importance(**params)
    append(f'feature-importance ({train_dir})')

""" genolearn evaluate """

def evaluate(train_dir):
    """  Given a training directory, evaluates a model on user prompted inputs and outputs to the evaluate subdirectory within the working directory """
    print(f'{PRE}\n\nCommand: evaluate\n\nEvaluate parameters for "{train_dir}"\n')

    path = os.path.join(working_directory, 'train', train_dir)
    log  = read_log(os.path.join(path, 'train.log'))
    meta = log['meta']

    with open(os.path.join(working_directory, 'meta', meta)) as f:
        meta   = json.load(f)
        if set(meta['group']) == {'Train', 'Test'}:
            groups = []
        else:
            groups = list(meta['group'])
        groups += ['Train' ,'Test', 'unlabelled']

    info   = dict(output    = dict(prompt = 'output filename', type = click.Path()),
                  values    = dict(prompt = 'group values', type = click.Choice(groups), multiple = True))

    params = dict(train_dir = train_dir)
    params.update(prompt(info))

    log = read_log(os.path.join(path, 'train.log'))
    for key in ['meta', 'feature_selection']:
        params[key] = log[key]

    log = read_log(os.path.join(path, 'params.json'))
    params['num_features'] = log['num_features']

    print_dict('executing "evaluate" with parameters:', params)

    log = read_log(os.path.join(path, 'encoding.json'))
    params['encoder'] = log

    path   = os.path.join(path, 'evaluate')

    os.makedirs(path, exist_ok = True)

    params['output'] = os.path.join(path, params['output'])

    if not params['output'].endswith('.csv'):
        params['output'] = params['output'] + '.csv'

    from   genolearn.core.evaluate import evaluate

    data_config = dict(working_dir = working_directory, meta_file = params.pop('meta'))

    params['data_config'] = data_config

    params['model'] = os.path.join(os.path.join(working_directory, 'train', params.pop('train_dir')), 'model.pickle')

    os.chdir(os.path.dirname(path))

    evaluate(**params)
    append(f'evaluate ({train_dir} {os.path.basename(params["output"]).replace(".csv", "")})')

def menu():
    """ Main menu for GenoLearn """

    _feature_importance = select_train_dir('feature-importance', feature_importance, menu, detect_feature_importance)
    _evaluate           = select_train_dir('evaluate', evaluate, menu, detect_evaluate)

    options             = {'print'              : {'info' : 'prints various GenoLearn generated files',
                                                   'func' : _print},
                           'preprocess'         : {'info' : 'preprocess data into an easier format for file reading',
                                                   'func' : preprocess},
                           'feature-selection'  : {'info' : 'computes a feature selection method for later training',
                                                   'func' : _feature_selection},
                           'model-config'       : {'info' : 'creates a machine learning model config',
                                                   'func' : model_config},
                           'train'              : {'info' : 'trains a machine learning model',
                                                   'func' : _train},
                           'feature-importance' : {'info' : 'computes the model feature importances',
                                                   'func' : _feature_importance},
                           'evaluate'           : {'info' : 'evaluates a trained model on an input dataset',
                                                   'func' : _evaluate}}

    # set k to not truncate options initially
    k = None

    # set k = 1 if not in working directory or has not executed setup
    if not check_working_directory():
        return print('unknown working directory - cd into the working directory and execute genolearn again or execute ' \
                     'genolearn-setup to setup this directory as the working directory')
        

    # if a command has not been previously executed truncate options to not include other commands that rely on it.
    else:
        for i, command in enumerate(['preprocess meta', 'feature-selection', 'model', 'train'], 2):
            if not check_history(command):
                k = i
                break
    try:
        enum(options, '', k = k, back = exit)
    except KeyboardInterrupt:
        print()
    except click.exceptions.Abort:
        print()


from   genolearn.logger     import Writing

from   sklearn.ensemble     import AdaBoostClassifier, RandomForestClassifier
from   sklearn.linear_model import LogisticRegression

import numpy as np
import os

class FeatureImportance():
    """
    Feature Importance Class

    Parameters
    ----------
        model : object
            Fitted scikit-learn model.

    Attributes
    ----------

        feature_score : numpy.ndarray
            Numpy array (dtype = float) of feature importance scores.

    Methods
    -------

        ranked_features: numpy.ndarray
            Numpy array (dtype = int) of ranked features.
    """
    def __init__(self, model):
        self.model = model
        self._score, self._rank = _assign(model)
        
    @property
    def feature_scores(self):
        return self._score(self.model)

    def feature_ranks(self, **kwargs):
        return self._rank(self.model, **kwargs)

    def save(self, name, path, **kwargs):
        fullpath = os.path.join(path, name)
        with Writing(fullpath, inline = True):
            np.savez(fullpath, feature_scores = self.feature_scores, feature_ranks = self.feature_ranks(**kwargs))


def _assign(model):
    if isinstance(model, LogisticRegression):
        return _logistic_score, _logistic_rank
    elif isinstance(model, (RandomForestClassifier, AdaBoostClassifier)):
        return _ensemble_score, _ensemble_rank
    else:
        raise Exception()

def _logistic_score(model):
    return model.coef_

def _logistic_rank(model, fun = np.absolute, **kwargs):
    return fun(_logistic_score(model)).mean(axis = -1).argsort()[::-1]

def _ensemble_score(model):
    return model.feature_importances_

def _ensemble_rank(model, **kwargs):
    return _ensemble_score(model).argsort()[::-1]

def feature_importance(meta, feature_selection, model, output):

    from   genolearn.dataloader import DataLoader
    from   genolearn.logger     import Writing, msg

    import pickle

    import pandas as pd
    import numpy  as np

    dataloader = DataLoader(meta)

    selection  = dataloader.load_feature_selection(feature_selection).argsort()

    features   = dataloader.features(selection)

    with open(model, 'rb') as f:
        model = pickle.load(f)

    importance = FeatureImportance(model)

    scores     = importance.feature_scores
    ranks      = importance.feature_ranks()

    features   = np.array(features)[ranks]

    npz        = os.path.join(output, 'importance.npz')
    csv        = os.path.join(output, 'importance-rank.csv')

    os.makedirs(output, exist_ok = True)

    with Writing(npz, inline = True):
        np.savez(npz, features = features, ranks = ranks, scores = scores)

    with Writing(csv, inline = True):
        with open(csv, 'w') as f:
            f.write('\n'.join(features[ranks]))

    msg(f'executed "genolearn feature-importance"')


"""
logger doc
"""

from   multiprocessing import Value, Lock

import logging
import sys
import time
import os

clear = '\033[2K\033[1G'
up    = f'{clear}\033[F'
space = '    '
logging.basicConfig(stream = sys.stdout, level = logging.INFO, format = f'{clear}%(asctime)s %(message)s', datefmt = "[%d-%m-%Y %H:%M:%S]")
logging.getLogger().handlers[0].terminator = ''
for name in logging.root.manager.loggerDict:
    logging.getLogger(name).setLevel(logging.CRITICAL)
    
def msg(text, inline = False, delete = 0, indent = 0):
    """
    Verbose message to print.

    Args:
        text (str)    : Message to print to sys.stdout.
        inline (bool) : If True, prints over the current line.
        delete (int)  : Number of previous lines to delete before printing.
        indent (int)  : Number of indentations (4 spaces) to indent text message by.
    """
    text = f'{space * indent}{text}' + ('' if inline else '\n')
    for _ in range(delete):
        logging.info(up)
    logging.info(text)


def print_dict(text, d):
    """
    Prints contents of dictionary using the msg function.

    Args:
        d (dict) : Dictionary of parameters. 
    """
    from  genolearn.utils   import set_params
    set_params(d)
    m    = max(len(key) for key in d)
    sep  = f'\n{clear}{" " * 23}'
    end  = f'\n{clear}{" " * 22}'
    text = f'{text}{end}' + '{' + f'{sep}' + f'{sep}'.join([f'{key:{m}s} : {val}' for key, val in d.items()]) + f'{end}' + '}'
    msg(text)


def _eta(seconds):
    """
    Computes eta in [hours,] minutes, seconds.

    Args:
        seconds (float): Time in seconds.

    Returns:
        str: [hours,] minutes, seconds
    """
    minutes, seconds = divmod(seconds, 60)
    hours  , minutes = divmod(minutes, 60)
    if hours:
        return f'{hours:.0f}h {minutes:02.0f}m {seconds:02.0f}s'
    return f'{minutes:02.0f}m {seconds:02.0f}s'

class Update():
    """
    Update class designed for verbose print outs in loops.
    
    Args:
        obj (object) : Object to iterate over.
        msg (str)    : Base string to print out during object iteration.
        final (str)  : Final string to print out once object has been iterated.
        eta (bool)   : If True, includes estimated eta to print outs.
    """
    def __init__(self, obj, msg = '', final = None, eta = False):
        self.obj   = iter(obj)
        self.msg   = msg
        self._msg  = ''
        self.num   = len(obj)
        self.i     = 0
        self.final = final
        self.eta   = eta
        self.time  = time.time()

    def compute_msg(self):
        per     = self.i / self.num
        msg     = f'{self.msg} {self.i:,d} of {self.num:,d} ({per:.2%})'
        if self.eta and (self.num > self.i > 1):
            delta = time.time() - self.time
            eta   = delta / self.i * (self.num - self.i)
            msg   = f'{msg} | eta {_eta(eta)}'
        self._msg = msg

    def __iter__(self):
        return self

    def __next__(self):
        final = (self.num == self.i)
        if final:
            if self.final:
                msg(self.final)
            raise StopIteration
        self.i += 1
        self.compute_msg()
        msg(self._msg, inline = True)
        return next(self.obj)

class Waiting():

    def __init__(self, during, after, file = None, inline = True, delete = 0):
        self.during  = during
        self.after   = after
        self.file    = file if file and ' ' in file else f'"{os.path.basename(file)}"' if file else ''
        self.inline  = inline
        self.delete  = delete

    @property
    def _during(self):
        return f'{self.during} {self.file}' if self.file else self.during
    
    @property
    def _after(self):
        return f'{self.after} {self.file}' if self.file else self.after

    def __enter__(self):
        msg(f'{self._during}', inline = self.inline)

    def __exit__(self, *args):
        msg(f'{self._after}', delete = self.delete)

class Executing(Waiting):

    def __init__(self, file, inline = False, delete = 0):
        super().__init__('executing', 'executed', file = file, inline = inline, delete = delete)

class Computing(Waiting):

    def __init__(self, file = None, inline = False, delete = 0):
        super().__init__('computing', 'computed', file = file, inline = inline, delete = delete)

class Writing(Waiting):

    def __init__(self, file = None, inline = False, delete = 0):
        super().__init__('writing', 'written', file = file, inline = inline, delete = delete)

class Counter():

    def __init__(self, message, verbose = 0, display = True):
        self.val      = Value('i', 0)
        self.lock     = Lock()
        self.verbose  = verbose
        self.message  = message
        self.display  = display
        
    def increment(self):
        with self.lock:
            self.val.value += 1
            if self.display: msg(self.message.format(self.val.value), inline = self.val.value % self.verbose if self.verbose else True)

    def value(self):
        with self.lock:
            return self.val.value


import shutil
import numpy as np
import os
import re

clean_sample   = lambda sample : sample.replace('/', '.')
gather_feature = lambda line : line[:line.index(' ')]
gather_samples = lambda line : list(map(clean_sample, re.findall(r'(?<= )[^: ]+(?=:)', line)))
gather_counts  = lambda line : re.findall(r'(?<=:)[0-9]+', line)

def get_dtype(val):
    dtypes = [np.uint8, np.uint16, np.uint32, np.uint64]
    for dtype in dtypes:
        info = np.iinfo(dtype)
        if info.min <= val <= info.max:
            return dtype
    raise Exception()

def init(file, subpath = 'temp', ext = 'txt'):
    path = os.path.join(subpath, f'{file}.{ext}') if subpath else f'{file}.{ext}'
    if os.path.exists(path):
        os.remove(path)
    return open(path, 'a')

def add(object, i, count):
    object.write(f'{i} {count}\n')
    
def preprocess(preprocess_dir, data, batch_size, n_processes, sparse, dense, max_features, verbose):
    """
    Preprocess a gunzip (gz) compressed text file containing genome sequence data of the following sparse format

        \b
        sequence_1 | identifier_{1,1}:count_{1,1} identifier_{1,1}:count_{2,1} ...
        sequence_2 | identifier_{2,1}:count_{2,1} identifier_{2,1}:count_{2,2} ...
        ...

    into a directory of .npz files, a list of all the features, and some meta information containing number of 
    identifiers, sequences, and non-zero counts.

    It is expected that the parameter `data` is in the `data_dir` directory set in the \033[1mactive config\033[0m file.
    See https://genolearn.readthedocs.io/tutorial/config for more details.

    \b\n
    Example Usage

    \b
    # reduce ram usage by setting a batch size (will increase preprocess time)
    >>> genolearn preprocess file.gz --batch-size 128
    """

    from   genolearn.logger       import msg, Waiting, print_dict
    from   genolearn              import utils

    from   pathos.multiprocessing import cpu_count, Pool
    from   shutil                 import rmtree

    import resource

    import json
    import gzip

    try:
        # remove number of files open restriction
        limit = min(resource.getrlimit(resource.RLIMIT_NOFILE)[1], batch_size + 100)
        resource.setrlimit(resource.RLIMIT_NOFILE, (limit, limit))
    except:
        msg(f'Attempting to open to many files! Try reducing the number')
        return 

    if max_features == None:
        max_features = -1

    n_processes    = cpu_count() if n_processes == 'auto' else int(n_processes)
    
    first_run  = True
    features   = []
    exceptions = set()
    C          = 0
    hi         = 0
    unique     = set()

    if data.endswith('.gz'):
        _open  = gzip.GzipFile
        decode = lambda line : line.decode()
    else:
        _open  = open
        decode = lambda line : line

    with _open(os.path.expanduser(data)) as gz:
        
        if os.path.exists(preprocess_dir):                    
            rmtree(preprocess_dir)

        os.mkdir(preprocess_dir)
        os.chdir(preprocess_dir)

        if 'temp' in os.listdir():
            shutil.rmtree('temp')

        os.mkdir('temp')

        files = {}

        while True:
           
            gz.seek(0)

            skip    = False
            skipped = False
            c       = 0
           
            for m, line in enumerate(gz, 1):

                line   = decode(line)

                srrs   = gather_samples(line)
                counts = gather_counts(line)

                if first_run:
                    features.append(gather_feature(line))
                    hi      = max(hi, *map(int, counts))
                    unique |= set(srrs)

                for SRR, count in zip(srrs, counts):
                    if SRR not in exceptions:
                        if SRR not in files and c <= batch_size:
                            if skip:
                                skipped = True
                                continue
                            files[SRR] = init(SRR)
                            c         += 1
                            C         += 1
                            skip = c == batch_size
                        add(files[SRR], m - 1, count)

                msg(f'{C:10,d} {m:10,d}', inline = m % verbose)                  
               
                if m == max_features:
                    break

            for f in files.values():
                f.close()

            if first_run:
                first_run = False

                d_dtype   = get_dtype(hi)
                c_dtype   = get_dtype(m)

                with Waiting('compressing', 'compressed', 'features.txt.gz'):
                    with gzip.open('features.txt.gz', 'wb') as g:
                        g.write(' '.join(features).encode())
                
                features.clear()

                f = init('info', None, 'json')
                json.dump({'n' : len(unique), 'm' : m, 'max' : hi}, f)
                f.close()
               
                def to_sparse(npz, c, d):
                    np.savez_compressed(os.path.join('sparse', npz), col = c.astype(c_dtype), data = d.astype(d_dtype))

                def to_dense(npz, c, d):
                    arr = np.zeros(m, dtype = d.dtype)
                    arr[c] = d
                    np.savez_compressed(os.path.join('dense', npz), arr = arr)
               
                def convert(file):
                    txt  = os.path.join('temp', f'{file}.txt')
                    npz  = f'{file}.npz'
                    c, d = np.loadtxt(txt, dtype = c_dtype).T

                    for function in functions:
                        function(npz, c, d)

                    os.remove(txt)

                functions = []
                if sparse:
                    functions.append(to_sparse)
                    os.mkdir('sparse')

                if dense:
                    functions.append(to_dense)
                    os.mkdir('dense')
           
            with Waiting('converting', 'converted', 'to arrays'):
                with Pool(n_processes) as pool:
                    pool.map(convert, list(files))

            if not skipped:
                break

            exceptions |= set(files)

            if len(files) < batch_size:
                break

            files.clear()

        os.rmdir('temp')
        
        utils.create_log('preprocess')

        os.chdir('..')

    msg(f'executed "preprocess sequence"')

def combine(preprocess_dir, data, batch_size, n_processes, max_features, verbose):
    """
    Preprocess a gunzip (gz) compressed text file containing genome sequence data of the following sparse format

        \b
        sequence_1 | identifier_{1,1}:count_{1,1} identifier_{1,1}:count_{2,1} ...
        sequence_2 | identifier_{2,1}:count_{2,1} identifier_{2,1}:count_{2,2} ...
        ...

    and combines the preprocessed data with the `preprocess_dir` directory set in the \033[1mactive config\033[0m file.
    This relies on the user to have previously executed `genolearn preprocess`.

    See https://genolearn.readthedocs.io/tutorial/combine for more details.
    """

    from   genolearn.logger       import msg, Waiting
    from   genolearn              import utils
    from   pathos.multiprocessing import cpu_count, Pool

    import numpy  as np

    import resource

    import json
    import gzip
    import re
    import os

    assert os.path.exists(preprocess_dir)

    try:
        # remove number of files open restriction
        limit = min(resource.getrlimit(resource.RLIMIT_NOFILE)[1], batch_size + 100)
        resource.setrlimit(resource.RLIMIT_NOFILE, (limit, limit))
    except:
        msg(f'Attempting to open to many files! Try reducing the number')
        return 

    with gzip.open(os.path.join(preprocess_dir, 'features.txt.gz')) as gz:
        feature_set = gz.read().decode().split()
        
    n_processes    = cpu_count() if n_processes == 'auto' else int(n_processes)

    gather_feature = lambda line : line[:line.index(' ')]
    gather_samples = lambda line : re.findall(r'[\w]+(?=:)', line)
    gather_counts  = lambda line : re.findall(r'(?<=:)[\w]+', line)

    first_run  = True
    features   = []
    C          = 0
    hi         = 0
    unique     = set()

    if data.endswith('.gz'):
        _open  = gzip.GzipFile
        decode = lambda line : line.decode()
    else:
        _open  = open
        decode = lambda line : line

    with _open(os.path.expanduser(data)) as gz:
       
        os.chdir(preprocess_dir)
        
        if 'temp' in os.listdir():
            shutil.rmtree('temp')
        os.mkdir('temp')
        
        files      = {}
        exceptions = set([file.replace('.npz', '') for file in os.listdir('dense' if 'dense' in os.listdir() else 'sparse')])
        while True:
           
            gz.seek(0)

            skip    = False
            skipped = False
            c       = 0
           
            for m, line in enumerate(gz, 1):

                line    = decode(line)

                srrs    = gather_samples(line)
                counts  = gather_counts(line)

                if first_run:
                    features.append(gather_feature(line))
                    hi      = max(hi, *map(int, counts))
                    unique |= set(srrs)

                for SRR, count in zip(srrs, counts):
                    if SRR not in exceptions:
                        if SRR not in files:
                            if skip:
                                skipped = True
                                continue
                            files[SRR] = init(SRR)
                            c         += 1
                            C         += 1
                            skip = c == batch_size
                        add(files[SRR], m - 1, count)

                msg(f'{C:10,d} {m:10,d}', inline = m % verbose)
               
                if m == max_features:
                    break

            for f in files.values():
                f.close()

            if first_run:
                first_run = False

                n         = len(unique)
                d_dtype   = get_dtype(hi)
                c_dtype   = get_dtype(m)
                
                with open('info.json') as f:
                    meta = json.load(f)
                    n    = meta['n'] + n
                    m    = meta['m']
                    hi   = max(meta['max'], hi)

                f = init('meta', None, 'json')
                json.dump({'n' : n, 'm' : m, 'max' : hi}, f)
                f.close()
                
                feature_overlap = np.nonzero(np.isin(features, feature_set, assume_unique = True))[0]

                def to_sparse(npz, c, d):
                    np.savez_compressed(os.path.join('sparse', npz), col = c.astype(c_dtype), data = d.astype(d_dtype))

                def to_dense(npz, c, d):
                    arr = np.zeros(m, dtype = d.dtype)
                    arr[c] = d
                    np.savez_compressed(os.path.join('dense', npz), arr = arr)
               
                def convert(file):
                    txt  = os.path.join('temp', f'{file}.txt')
                    npz  = f'{file}.npz'
                    c, d = np.loadtxt(txt, dtype = c_dtype).T
                    mask = np.nonzero(np.isin(c, feature_overlap, assume_unique = True))[0]
                    for function in functions:
                        function(npz, c[mask], d[mask])

                    os.remove(txt)

                functions = []
                if 'sparse' in os.listdir():
                    functions.append(to_sparse)

                if 'dense' in os.listdir():
                    functions.append(to_dense)

            with Waiting('converting', 'converted', 'to arrays'):
                with Pool(n_processes) as pool:
                    pool.map(convert, list(files))

            if not skipped:
                break

            exceptions |= set(files)

            if len(files) < batch_size:
                break

            files.clear()

        os.rmdir('temp')

        utils.create_log('combine')

        os.chdir('..')

    msg(f'executed "preprocess combine"')


def preprocess_meta(output, meta_path, identifier_column, target_column, group_column, train_values, test_values, ptrain):

    from   genolearn import get_active

    import pandas as pd
    import json

    active   = get_active()

    pdir     = 'preprocess'
    file_dir = os.path.join(pdir, 'sparse' if 'sparse' in os.listdir(pdir) else 'dense')
    files    = [file.replace('.npz', '') for file in os.listdir(file_dir)]
    meta_df  = pd.read_csv(meta_path).applymap(str)
    meta_df[identifier_column] = meta_df[identifier_column].apply(clean_sample)
    
    valid    = set(files)

    meta_df  = meta_df.loc[meta_df[identifier_column].isin(valid)]

    if group_column == 'None':
        group_column                      = 'train_test'
        n                                 = len(meta_df)
        i                                 = int(n * ptrain + 0.5)
        r                                 = np.random.permutation(n)
        values                            = np.array(['Train'] * n)
        values[r[i:]]                     = 'Test'
        meta_df[group_column]             = values
        
    groups    = sorted(set(meta_df[group_column]))

    meta_json = {'identifiers' : list(meta_df[identifier_column]), 'targets' : sorted(set(meta_df[target_column]))}

    meta_json['search'] = {}
    meta_json['group']  = {}
    for i, row in meta_df.iterrows():
        meta_json['search'][row[identifier_column]] = row[target_column]

    for group in groups:
        meta_json['group'][group] = list(meta_df.loc[meta_df[group_column] == group, identifier_column])

    meta_json['Train'] = []
    meta_json['Test' ] = []

    for group in groups:
        if group in train_values:
            meta_json['Train'].append(group)
        elif group in test_values:
            meta_json['Test'].append(group)
    
    os.makedirs('meta', exist_ok = True)

    with open(os.path.join('meta', output), 'w') as file:
        print(json.dumps(meta_json, indent = 4), file = file)

    print(f'created "{output}" in meta')



from   genolearn.models  import classification
from   genolearn.logger  import Waiting, msg
from   genolearn.utils   import monitor_RAM
from   genolearn.metrics import Metrics
from   itertools         import product
from   time              import time

import numpy as np
import joblib
import os

root = 'models'

def set_dir(path):
    """test"""
    global root
    root = path

def get_dir():
    global root
    return root

def save(model, path, overwrite = False):
    global root
    os.makedirs(root, exist_ok = True)
    full_path = os.path.join(root, path)
    if os.path.exists(full_path):
        if overwrite:
            raise Exception(f'"{full_path}" already exists!')
    
    joblib.dump(model, full_path)

def load(path):
    global root
    full_path = os.path.join(root, path)
    if os.path.exists(full_path):
        return joblib.load(path)
    raise Exception(f'"{full_path}" does not exist!')

def grid_predictions(dataloader, Model, selection, num_features, common_kwargs = {}, min_count = 0, target_subset = None, metric = 'recall', mean_func = 'weighted_mean', **kwargs):

    values  = [num_features] + list(kwargs.values())
    params  = list(product(*values))
    names   = ['num_features'] + list(kwargs)
    C       = [len(num_features)] + [len(values) for values in kwargs.values()]
    M       = len(C)
    V       = [None] * len(names)
    
    with Waiting('loading', 'loaded', 'train / test data', inline = True):
        X_train, Y_train, X_test, Y_test = dataloader.load_train_test(features = selection[:max(num_features)], min_count = min_count, target_subset = target_subset)

    keys    = ['predict', 'predict_proba', 'predict_log_proba']
    outputs = {'target' : dataloader.decode(Y_test), 'labels' : list(dataloader._encoder), 'time' : [], 'num_features' : num_features}
    for key in keys:
        if hasattr(Model, key):
            outputs[key] = []

    best = (None, None, -1)
    for k, param in enumerate(params):
        for i, theta in enumerate(param):
            flag = V[i] != theta
            if (i + 1) < M and V[i] is not None and V[i] != theta:
                flag   = True
                delete = sum(C[i + 1:])
                for j in range(i + 1, M):
                    V[j] = None
            else:
                delete = 0
        
            V[i]    = theta
            message = ' '.join([f'{p} {v}' for p, v in zip(names[:i + 1], V[:i + 1])])

            if flag:
                msg(message, delete = delete)
        
        model = Model(**common_kwargs, **dict(zip(names[1:], V[1:])))

        start = time()
        model.fit(X_train[:,:param[0]], Y_train)
        fit   = time()
        hat   = model.predict(X_test[:,:param[0]])
        pred  = time()

        outputs['time'].append((fit - start, pred - fit))

        for key in keys:
            if hasattr(model, key):
                outputs[key].append(getattr(model, key)(X_test[:,:param[0]]))

        score = Metrics(Y_test, hat, metric)(func = mean_func)
        if score > best[2]:
            best_kwargs = {**common_kwargs, **dict(zip(names[1:], V[1:])), 'num_features' : param[0]}
            best        = (model, k, score)

        monitor_RAM()
    
    k                      = best[1]
    best                   = (best[0], dataloader.decode(outputs['predict'][k]))

    if hasattr(Model, 'predict_proba'):
        best              += (outputs['predict_proba'][k],)

    outputs['identifiers'] = dataloader.identifiers_test
    outputs['predict']     = dataloader.decode(np.array(outputs['predict']).reshape(*C, -1))
    
    for key in keys[1:]:
        if hasattr(Model, key):
            outputs[key]   = np.array(outputs[key]).reshape(*C, -1, len(dataloader._encoder))

    outputs['time']        = np.array(outputs['time']).reshape(*C, 2)
        
    outputs.update(common_kwargs)
    outputs.update(kwargs)

    outputs['best']   = best

    msg('computed predictions and computation times', delete = sum(C))
    
    return outputs, best_kwargs